/*
 * File: window.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 * MACOS port by Aria Stewart <aredridel@nbtsc.org>
 */


/***

Window management. Some of the functions are internal, and some are
attached to keys that the user actually types.

***/

#include	<stdio.h>
#include	<stdlib.h>
#include	"rho.h"
#if MACOS
#include	<termios.h>
#include        <util.h>
#else
#include	<asm/termios.h>
#include	<asm/ioctls.h>
#endif                          /* MACOS */
#include	"window.h"
#include	"cursor.h"
#include	"display.h"
#include	"messages.h"

extern int ioctl(int, int, void *);

/*++

Put window in VIEW mode and update mode lines

++*/
void window_ro()
{
    curwp->w_bufp->b_mode |= MDVIEW;
    update_modeline();

}

/*++

Helper for check_resize which itself is used by scopen

++*/
static int newsize(f, n)
int f, n;

{
    WNDW *wp;                   /* current window being examined */
    WNDW *nextwp;               /* next window to scan */
    WNDW *lastwp;               /* last window scanned */
    int lastline;               /* screen line of last line of current window */
    int cmark;                  /* current mark */

    /* if the command defaults, assume the largest */
    if (f == FALSE)
        n = term.t_mrow + 1;

    /* make sure it's in range */
    if (n < 3 || n > term.t_mrow + 1) {
        mlwrite(TEXT209);       /*"%%Screen size out of range" */
        return (FALSE);
    }

    if (term.t_nrow == n - 1)
        return (TRUE);
    else if (term.t_nrow < n - 1) {

        /* go to the last window */
        wp = wheadp;
        while (wp->w_wndp != NULL)
            wp = wp->w_wndp;

        /* and enlarge it as needed */
        wp->w_ntrows = n - wp->w_toprow - 2;
        wp->w_flag |= WFHARD | WFMODE;

    } else {

        /* rebuild the window structure */
        nextwp = wheadp;
        wp = NULL;
        lastwp = NULL;
        while (nextwp != NULL) {
            wp = nextwp;
            nextwp = wp->w_wndp;

            /* get rid of it if it is too low */
            if (wp->w_toprow > n - 2) {

                /* save the point/mark if needed */
                if (--wp->w_bufp->b_nwnd == 0) {
                    wp->w_bufp->b_dotp = wp->w_dotp;
                    wp->w_bufp->b_doto = wp->w_doto;
                    for (cmark = 0; cmark < NMARKS; cmark++) {
                        wp->w_bufp->b_markp[cmark] = wp->w_markp[cmark];
                        wp->w_bufp->b_marko[cmark] = wp->w_marko[cmark];
                    }
                    wp->w_bufp->b_fcol = wp->w_fcol;
                }

                /* update curwp and lastwp if needed */
                if (wp == curwp)
                    curwp = wheadp;
                curbp = curwp->w_bufp;
                if (lastwp != NULL)
                    lastwp->w_wndp = NULL;

                /* free the structure */
                free((char *) wp);
                wp = NULL;

            } else {
                /* need to change this window size? */
                lastline = wp->w_toprow + wp->w_ntrows - 1;
                if (lastline >= n - 2) {
                    wp->w_ntrows = n - wp->w_toprow - 2;
                    wp->w_flag |= WFHARD | WFMODE;
                }
            }

            lastwp = wp;
        }
    }

    /* screen is garbage */
    term.t_nrow = n - 1;
    sgarbf = TRUE;
    return (TRUE);
}



/*++

Helper for check_resize which itself is used by scopen

++*/
static int newwidth(f, n)
int f, n;

{
    WNDW *wp;

    /* if the command defaults, assume the largest */
    if (f == FALSE)
        n = term.t_mcol;

    /* make sure it's in range */
    if (n < 10 || n > term.t_mcol) {
        mlwrite(TEXT210);       /*"%%Screen width out of range" */
        return (FALSE);
    }

    /* otherwise, just re-width it (no big deal) */
    term.t_ncol = n;
    term.t_margin = n / 10;
    term.t_scrsiz = n - (term.t_margin * 2);

    /* force all windows to redraw */
    wp = wheadp;
    while (wp) {
        wp->w_flag |= WFHARD | WFMOVE | WFMODE;
        wp = wp->w_wndp;
    }
    sgarbf = TRUE;

    return (TRUE);
}



/*++

Helper for unix.scopen

++*/
void check_resize(sig)
int sig;
{
    int row, col;
    struct winsize winsz;

    sig = 0;                    /* splint */

    ioctl(0, TIOCGWINSZ, &winsz);
    col = (int) winsz.ws_col;
    row = (int) winsz.ws_row;

    if (col > 256)
        col = 256;
    if (row > 120)
        row = 120;

    if ((col != term.t_mcol) || (row != term.t_mrow)) {
        term.t_mcol = col;
        term.t_mrow = row;
        if (term.t_nrow != term.t_mrow - 1) {   /* Center dot. */
            curwp->w_force = 0;
            curwp->w_flag |= WFFORCE;
            newsize(TRUE, term.t_mrow);
        }
        if (term.t_ncol != term.t_mcol) {       /* Center dot. */
            curwp->w_force = 0;
            curwp->w_flag |= WFFORCE;
            newwidth(TRUE, term.t_mcol);
        }
        return;
    }
    return;
}



/*++

Refresh the screen. With no argument, it just does the refresh. With an
argument it recenters "." in the current window. Bound to "C-L".

++*/
int refresh_scrn(f, n)
int f, n;

{
    n = 0;                      /* splint */

    if (f == FALSE)
        sgarbf = TRUE;
    else {
        curwp->w_force = 0;     /* Center dot. */
        curwp->w_flag |= WFFORCE;
    }

    return (TRUE);
}

/*++

The command make the next window (next => down the screen) the current
window. There are no real errors, although the command does nothing if
there is only 1 window on the screen.  With an argument this command
finds the <n>th window from the top

++*/
int next_window(f, n)
int f, n;

{
    WNDW *wp;
    int nwindows;               /* total number of windows */

    if (f) {

        /* first count the # of windows */
        wp = wheadp;
        nwindows = 1;
        while (wp->w_wndp != NULL) {
            nwindows++;
            wp = wp->w_wndp;
        }

        /* if the argument is negative, it is the nth window
           from the bottom of the screen                        */
        if (n < 0)
            n = nwindows + n + 1;

        /* if an argument, give them that window from the top */
        if (n > 0 && n <= nwindows) {
            wp = wheadp;
            while (--n)
                wp = wp->w_wndp;
        } else {
            mlwrite(TEXT203);   /* "Window number out of range" */
            return (FALSE);
        }
    } else if ((wp = curwp->w_wndp) == NULL)
        wp = wheadp;
    curwp = wp;
    curbp = wp->w_bufp;
    update_modeline();
    return (TRUE);
}



/*++

This command makes the previous window (previous => up the screen) the
current window. There aren't any errors, although the command does not do a
lot if there is 1 window.

++*/
int previous_window(f, n)
int f, n;

{
    WNDW *wp1;
    WNDW *wp2;

    /* if we have an argument, we mean the nth window from the bottom */
    if (f)
        return (next_window(f, -n));

    wp1 = wheadp;
    wp2 = curwp;

    if (wp1 == wp2)
        wp2 = NULL;

    while (wp1->w_wndp != wp2)
        wp1 = wp1->w_wndp;

    curwp = wp1;
    curbp = wp1->w_bufp;
    update_modeline();
    return (TRUE);
}



/*++

This command moves the current window down by "arg" lines. Recompute the
top line in the window. The move up and move down code is almost completely
the same; most of the work has to do with reframing the window, and picking
a new dot. We share the code by having "move down" just be an interface to
"move up". 

++*/
int window_down(f, n)
int f, n;

{
    return (window_up(f, -n));
}



/*++

Move the current window up by "arg" lines. Recompute the new top line of
the window. Look to see if "." is still on the screen. If it is, you win.
If it isn't, then move "." to center it in the new framing of the window
(this command does not really move "."; it moves the frame). Bound to
"C-X C-P".

++*/
int window_up(f, n)
int f, n;

{
    LINE *lp;
    int i;

    f = 0;                      /* splint */

    lp = curwp->w_linep;

    if (n < 0) {
        while (n++ && lp != curbp->b_linep)
            lp = lforw(lp);
    } else {
        while (n-- && lback(lp) != curbp->b_linep)
            lp = lback(lp);
    }

    curwp->w_linep = lp;
    curwp->w_flag |= WFHARD;    /* Mode line is OK. */

    for (i = 0; i < curwp->w_ntrows; ++i) {
        if (lp == curwp->w_dotp)
            return (TRUE);
        if (lp == curbp->b_linep)
            break;
        lp = lforw(lp);
    }

    lp = curwp->w_linep;
    i = curwp->w_ntrows / 2;

    while (i-- && lp != curbp->b_linep)
        lp = lforw(lp);

    curwp->w_dotp = lp;
    curwp->w_doto = 0;
    return (TRUE);
}



/*++

This command makes the current window the only window on the screen. Bound
to "C-X 1". Try to set the framing so that "." does not have to move on the
display. Some care has to be taken to keep the values of dot and mark in
the buffer structures right if the distruction of a window makes a buffer
become undisplayed.

++*/
int single_window(f, n)
int f, n;

{
    WNDW *wp;
    LINE *lp;
    int i;
    int cmark;                  /* current mark */

    f = n = 0;                  /* splint */

    while (wheadp != curwp) {
        wp = wheadp;
        first_screen->s_first_window = wheadp = wp->w_wndp;
        if (--wp->w_bufp->b_nwnd == 0) {
            wp->w_bufp->b_dotp = wp->w_dotp;
            wp->w_bufp->b_doto = wp->w_doto;
            for (cmark = 0; cmark < NMARKS; cmark++) {
                wp->w_bufp->b_markp[cmark] = wp->w_markp[cmark];
                wp->w_bufp->b_marko[cmark] = wp->w_marko[cmark];
            }
            wp->w_bufp->b_fcol = wp->w_fcol;
        }
        free((char *) wp);
    }
    while (curwp->w_wndp != NULL) {
        wp = curwp->w_wndp;
        curwp->w_wndp = wp->w_wndp;
        if (--wp->w_bufp->b_nwnd == 0) {
            wp->w_bufp->b_dotp = wp->w_dotp;
            wp->w_bufp->b_doto = wp->w_doto;
            for (cmark = 0; cmark < NMARKS; cmark++) {
                wp->w_bufp->b_markp[cmark] = wp->w_markp[cmark];
                wp->w_bufp->b_marko[cmark] = wp->w_marko[cmark];
            }
            wp->w_bufp->b_fcol = wp->w_fcol;
        }
        free((char *) wp);
    }
    lp = curwp->w_linep;
    i = curwp->w_toprow;
    while (i != 0 && lback(lp) != curbp->b_linep) {
        --i;
        lp = lback(lp);
    }
    curwp->w_toprow = 0;
    curwp->w_ntrows = term.t_nrow - 1;
    curwp->w_linep = lp;
    curwp->w_flag |= WFMODE | WFHARD;
    return (TRUE);
}



/*++

Delete the current window, placing its space in the window above,
or, if it is the top window, the window below. Bound to C-X 0.

++*/
int delete_window(f, n)
int f, n;

{
    WNDW *wp;                   /* window to recieve deleted space */
    WNDW *lwp;                  /* ptr window before curwp */
    int target;                 /* target line to search for */
    int cmark;                  /* current mark */

    f = n = 0;                  /* splint */

    /* if there is only one window, don't delete it */
    if (wheadp->w_wndp == NULL) {
        mlwrite(TEXT204);       /*"Can not delete this window" */
        return (FALSE);
    }

    /* find window before curwp in linked list */
    wp = wheadp;
    lwp = NULL;
    while (wp != NULL) {
        if (wp == curwp)
            break;
        lwp = wp;
        wp = wp->w_wndp;
    }

    /* find recieving window and give up our space */
    wp = wheadp;
    if (curwp->w_toprow == 0) {
        /* find the next window down */
        target = curwp->w_ntrows + 1;
        while (wp != NULL) {
            if (wp->w_toprow == target)
                break;
            wp = wp->w_wndp;
        }
        if (wp == NULL)
            return (FALSE);
        wp->w_toprow = 0;
        wp->w_ntrows += target;
    } else {
        /* find the next window up */
        target = curwp->w_toprow - 1;
        while (wp != NULL) {
            if ((wp->w_toprow + wp->w_ntrows) == target)
                break;
            wp = wp->w_wndp;
        }
        if (wp == NULL)
            return (FALSE);
        wp->w_ntrows += 1 + curwp->w_ntrows;
    }

    /* get rid of the current window */
    if (--curwp->w_bufp->b_nwnd == 0) {
        curwp->w_bufp->b_dotp = curwp->w_dotp;
        curwp->w_bufp->b_doto = curwp->w_doto;
        for (cmark = 0; cmark < NMARKS; cmark++) {
            curwp->w_bufp->b_markp[cmark] = curwp->w_markp[cmark];
            curwp->w_bufp->b_marko[cmark] = curwp->w_marko[cmark];
        }
        curwp->w_bufp->b_fcol = curwp->w_fcol;
    }
    if (lwp == NULL)
        first_screen->s_first_window = wheadp = curwp->w_wndp;
    else
        lwp->w_wndp = curwp->w_wndp;
    free((char *) curwp);
    curwp = wp;
    wp->w_flag |= WFHARD;
    curbp = wp->w_bufp;
    update_modeline();
    return (TRUE);
}

/*++

Split the current window.  A window smaller than 3 lines cannot be
split.	(Two line windows can be split when mode lines are disabled) An
argument of 1 forces the cursor into the upper window, an argument of
two forces the cursor to the lower window.  The only other error that
is possible is a "malloc" failure allocating the structure for the new
window.  Bound to "C-X 2". 

++*/

int split_window(f, n)
int f, n;

{
    WNDW *wp;
    LINE *lp;
    int ntru;
    int ntrl;
    int ntrd;
    WNDW *wp1;
    WNDW *wp2;
    int cmark;                  /* current mark */

    /* make sure we have enough space */
    if (curwp->w_ntrows < (modeflag ? 3 : 2)) {
        mlwrite(TEXT205, curwp->w_ntrows);      /*"Cannot split a %d line window" */
        return (FALSE);
    }
    wp = (WNDW *) malloc(sizeof(WNDW));


    ++curbp->b_nwnd;            /* Displayed twice.     */
    wp->w_bufp = curbp;
    wp->w_dotp = curwp->w_dotp;
    wp->w_doto = curwp->w_doto;
    for (cmark = 0; cmark < NMARKS; cmark++) {
        wp->w_markp[cmark] = curwp->w_markp[cmark];
        wp->w_marko[cmark] = curwp->w_marko[cmark];
    }
    wp->w_fcol = curwp->w_fcol;
    wp->w_flag = 0;
    wp->w_force = 0;

    if (n > 1)                  /* Upper size */
        ntru = n;
    else
        ntru = (curwp->w_ntrows - 1) / 2;

    ntrl = (curwp->w_ntrows - 1) - ntru;        /* Lower size */

    lp = curwp->w_linep;
    ntrd = 0;
    while (lp != curwp->w_dotp) {
        ++ntrd;
        lp = lforw(lp);
    }
    lp = curwp->w_linep;
    if (((f == FALSE) && (ntrd <= ntru)) || ((f == TRUE) && (n == 1))) {
        /* Old is upper window. */
        if (ntrd == ntru)       /* Hit mode line. */
            lp = lforw(lp);
        curwp->w_ntrows = ntru;
        wp->w_wndp = curwp->w_wndp;
        curwp->w_wndp = wp;
        wp->w_toprow = curwp->w_toprow + ntru + 1;
        wp->w_ntrows = ntrl;
    } else {
        /* Old is lower window  */
        wp1 = NULL;
        wp2 = wheadp;
        while (wp2 != curwp) {
            wp1 = wp2;
            wp2 = wp2->w_wndp;
        }
        if (wp1 == NULL)
            first_screen->s_first_window = wheadp = wp;
        else
            wp1->w_wndp = wp;
        wp->w_wndp = curwp;
        wp->w_toprow = curwp->w_toprow;
        wp->w_ntrows = ntru;
        ++ntru;                 /* Mode line.           */
        curwp->w_toprow += ntru;
        curwp->w_ntrows = ntrl;
        while (ntru--)
            lp = lforw(lp);
    }
    curwp->w_linep = lp;        /* Adjust the top lines */
    wp->w_linep = lp;           /* if necessary.        */
    curwp->w_flag |= WFMODE | WFHARD;
    wp->w_flag |= WFMODE | WFHARD;
    return (TRUE);
}


/*  circular refs in grow/shrinkwindow  */
static int growwindow(int, int);



/*++

Shrink the current window. Find the window that gains space. Hack at the
window descriptions. Ask the redisplay to do all the hard work. 

++*/
static int shrinkwindow(f, n)
int f, n;

{
    WNDW *adjwp;
    LINE *lp;
    int i;

    if (n < 0)
        return (growwindow(f, -n));
    if (wheadp->w_wndp == NULL) {
        mlwrite(TEXT206);       /*"Only one window" */
        return (FALSE);
    }
    if ((adjwp = curwp->w_wndp) == NULL) {
        adjwp = wheadp;
        while (adjwp->w_wndp != curwp)
            adjwp = adjwp->w_wndp;
    }
    if ((curwp->w_ntrows + (modeflag ? 0 : 1)) <= n) {
        mlwrite(TEXT207);       /*"Impossible change" */
        return (FALSE);
    }
    if (curwp->w_wndp == adjwp) {       /* Grow below.          */
        lp = adjwp->w_linep;
        for (i = 0; i < n && lback(lp) != adjwp->w_bufp->b_linep; ++i)
            lp = lback(lp);
        adjwp->w_linep = lp;
        adjwp->w_toprow -= n;
    } else {                    /* Grow above.          */
        lp = curwp->w_linep;
        for (i = 0; i < n && lp != curbp->b_linep; ++i)
            lp = lforw(lp);
        curwp->w_linep = lp;
        curwp->w_toprow += n;
    }
    curwp->w_ntrows -= n;
    adjwp->w_ntrows += n;
    curwp->w_flag |= WFMODE | WFHARD;
    adjwp->w_flag |= WFMODE | WFHARD;
    return (TRUE);
}



/*++

Enlarge the current window. Find the window that loses space. Make sure it
is big enough. If so, hack the window descriptions, and ask redisplay to do
all the hard work. You don't just set "force reframe" because dot would
move. 

++*/
static int growwindow(f, n)
int f, n;

{
    WNDW *adjwp;
    LINE *lp;
    int i;

    if (n < 0)
        return (shrinkwindow(f, -n));
    if (wheadp->w_wndp == NULL) {
        mlwrite(TEXT206);       /*"Only one window" */
        return (FALSE);
    }
    if ((adjwp = curwp->w_wndp) == NULL) {
        adjwp = wheadp;
        while (adjwp->w_wndp != curwp)
            adjwp = adjwp->w_wndp;
    }
    if ((adjwp->w_ntrows + (modeflag ? 0 : 1)) <= n) {
        mlwrite(TEXT207);       /*"Impossible change" */
        return (FALSE);
    }
    if (curwp->w_wndp == adjwp) {       /* Shrink below.        */
        lp = adjwp->w_linep;
        for (i = 0; i < n && lp != adjwp->w_bufp->b_linep; ++i)
            lp = lforw(lp);
        adjwp->w_linep = lp;
        adjwp->w_toprow += n;
    } else {                    /* Shrink above.        */
        lp = curwp->w_linep;
        for (i = 0; i < n && lback(lp) != curbp->b_linep; ++i)
            lp = lback(lp);
        curwp->w_linep = lp;
        curwp->w_toprow -= n;
    }
    curwp->w_ntrows += n;
    adjwp->w_ntrows -= n;
    curwp->w_flag |= WFMODE | WFHARD;
    adjwp->w_flag |= WFMODE | WFHARD;
    return (TRUE);
}



/*++

Resize the current window to the requested size.

++*/
int resize_window(f, n)
int f, n;

{
    int clines;                 /* current # of lines in window */

    /* must have a non-default argument, else ignore call */
    if (f == FALSE)
        return (TRUE);

    /* find out what to do */
    clines = curwp->w_ntrows;

    /* already the right size? */
    if (clines == n)
        return (TRUE);

    return (growwindow(TRUE, n - clines));
}



/*++

Pop up the indicated buffer

++*/
int wpopup(popbuf)
BUFFER *popbuf;                 /* content for popup */
{
    WNDW *wp;
    BUFFER *bp;
    int cmark;                  /* current mark */
    int height = POPUPS;        /* set in rho.h */

    /* on screen already? */
    if (popbuf->b_nwnd != 0)
        goto setwin;

    /* find the window to split */
    (void) single_window(0, 0);
    if (wheadp->w_wndp == NULL  /* Only 1 window        */
        && split_window(TRUE, height) == FALSE) /* and it won't split   */
        return (FALSE);
    wp = wheadp;                /* Find window to use   */
    while (wp != NULL && wp == curwp)
        wp = wp->w_wndp;

    if (popbuf->b_nwnd == 0) {  /* Not on screen yet.   */
        bp = wp->w_bufp;
        if (--bp->b_nwnd == 0) {
            bp->b_dotp = wp->w_dotp;
            bp->b_doto = wp->w_doto;
            for (cmark = 0; cmark < NMARKS; cmark++) {
                bp->b_markp[cmark] = wp->w_markp[cmark];
                bp->b_marko[cmark] = wp->w_marko[cmark];
            }
            bp->b_fcol = wp->w_fcol;
        }
        wp->w_bufp = popbuf;
        ++popbuf->b_nwnd;
    }

  setwin:wp = wheadp;
    while (wp != NULL) {
        if (wp->w_bufp == popbuf) {
            wp->w_linep = lforw(popbuf->b_linep);
            wp->w_dotp = lforw(popbuf->b_linep);
            wp->w_doto = 0;
            for (cmark = 0; cmark < NMARKS; cmark++) {
                wp->w_markp[cmark] = NULL;
                wp->w_marko[cmark] = 0;
            }
            wp->w_flag |= WFMODE | WFHARD;
            popbuf->b_mode |= MDVIEW;   /* put this buffer view mode */
            update_modeline();
        }
        wp = wp->w_wndp;
    }
    return (TRUE);
}



/*++

Scroll the next window up (back) a page

++*/
int scroll_next_up(f, n)
int f, n;

{

    f = n = 0;                  /* splint */

    next_window(FALSE, 1);
    page_up();
    previous_window(FALSE, 1);
    return (TRUE);
}


/*++

Scroll the next window down (forward) a page

++*/
int scroll_next_down(f, n)
int f, n;

{
    f = n = 0;                  /* splint */

    next_window(FALSE, 1);
    page_down();
    previous_window(FALSE, 1);
    return (TRUE);
}
